from gearpy.units import AngularPosition, AngularSpeed, AngularAcceleration, Angle, Force, InertiaMoment, Length, \
    Stress, Time, Torque, UnitBase
from math import sin, cos, sqrt
from .mechanical_object_base import RotatingObject, GearBase, lewis_factor_function, Role
from .mating_roles import MatingMaster, MatingSlave
from typing import Callable, Dict, List, Union, Optional


class SpurGear(GearBase):
    r"""``gearpy.mechanical_objects.spur_gear.SpurGear`` object.

    Attributes
    ----------
    :py:attr:`name` : str
        Name of the spur gear.
    :py:attr:`n_teeth` : int
        Number of gear teeth.
    :py:attr:`inertia_moment` : InertiaMoment
        Moment of inertia of the gear.
    :py:attr:`module` : Length
        Unit of the gear teeth size.
    :py:attr:`reference_diameter` : Length
        Reference diameter of the gear.
    :py:attr:`face_width` : Length
        Face width of the gear.
    :py:attr:`elastic_modulus` : Stress
        Elastic modulus of the material of the gear.
    :py:attr:`lewis_factor` : float
        Factor used to compute stresses on the gear tooth.
    :py:attr:`driven_by` : RotatingObject
        Rotating object that drives the gear, for example a motor, a flywheel or another gear.
    :py:attr:`drives` : RotatingObject
        Rotating object driven by the gear, it can be a flywheel or another gear.
    :py:attr:`master_gear_ratio` : float
        Gear ratio of the mating between the gear and its driving gear.
    :py:attr:`master_gear_efficiency` : float or int
        Efficiency of the gear mating between the gear and its driving gear.
    :py:attr:`mating_role`: Role
        The role of the gear in the gear mating.
    :py:attr:`angular_position` : AngularPosition
        Angular position of the gear.
    :py:attr:`angular_speed` : AngularSpeed
        Angular speed of the gear.
    :py:attr:`angular_acceleration` : AngularAcceleration
        Angular acceleration of the gear.
    :py:attr:`torque` : Torque
        Net torque applied on the gear.
    :py:attr:`driving_torque` : Torque
        Driving torque applied on the gear by its driving gear.
    :py:attr:`load_torque` : Torque
        Load torque applied on the gear by its driven gear or an external load.
    :py:attr:`tangential_force` : Force
        Tangential force applied on the gear teeth by the mating gear.
    :py:attr:`tangential_force_is_computable` : bool
        Whether is possible to compute the tangential force on the gear teeth.
    :py:attr:`bending_stress` : Stress
        Bending stress applied on the gear teeth by the mating gear.
    :py:attr:`bending_stress_is_computable` : bool
        Whether is possible to compute the bending stress on the gear teeth.
    :py:attr:`contact_stress` : Stress
        The stress generated by the contact with mating gear teeth.
    :py:attr:`contact_stress_is_computable` : bool
        Whether is possible to compute the contact stress on the gear teeth.
    :py:attr:`time_variables` : dict
        Time variables of the gear.

    Methods
    -------
    :py:meth:`compute_tangential_force`
        Computes the tangential force applied on the gear teeth by the mating gear.
    :py:meth:`compute_bending_stress`
        Computes the bending stress applied on the gear teeth by the mating gear.
    :py:meth:`compute_contact_stress`
        Computes the contact stress generated by the contact with mating gear teeth.
    :py:meth:`external_torque`
        Custom function to compute the external torque applied on the gear.
    :py:meth:`update_time_variables`
        Updates ``time_variables`` dictionary by appending the last value of each time variable to corresponding list.
    """

    def __init__(self,
                 name: str,
                 n_teeth: int,
                 inertia_moment: InertiaMoment,
                 module: Optional[Length] = None,
                 face_width: Optional[Length] = None,
                 elastic_modulus: Optional[Stress] = None):
        super().__init__(name = name,
                         n_teeth = n_teeth,
                         module = module,
                         face_width = face_width,
                         inertia_moment = inertia_moment,
                         elastic_modulus = elastic_modulus)

        if self.tangential_force_is_computable:
            self.time_variables['tangential force'] = []

            if self.bending_stress_is_computable:
                self.time_variables['bending stress'] = []
                self.__lewis_factor = lewis_factor_function(self.n_teeth).take(0)

                if self.contact_stress_is_computable:
                    self.time_variables['contact stress'] = []
                    self.__PRESSURE_ANGLE = Angle(20, 'deg')

    @property
    def name(self) -> str:
        """Name of the spur gear. It must be a non-empty string. \n
        It must be a unique name, not shared by other elements in the powertrain elements. \n
        Once set at the spur gear instantiation, it cannot be changed afterward.

        Returns
        -------
        str
            Name of the spur gear.

        Raises
        ------
        TypeError
            If ``name`` is not a string.
        ValueError
            If ``name`` is an empty string.
        """
        return super().name

    @property
    def n_teeth(self) -> int:
        """Number of gear teeth. It must be a positive integer. \n
        Once set at the spur gear instantiation, it cannot be changed afterward.

        Returns
        -------
        int
            Number of gear teeth.

        Raises
        ------
        TypeError
            If ``n_teeth`` is not an integer.
        ValueError
            If ``n_teeth`` is less than the minimum number of teeth, based on Lewis Factor table.
        """
        return super().n_teeth

    @property
    def inertia_moment(self) -> InertiaMoment:
        """Moment of inertia of the gear. It must be an instance of ``InertiaMoment``. \n
        Once set at the spur gear instantiation, it cannot be changed afterward.

        Returns
        -------
        InertiaMoment
            Moment of inertia of the gear.

        Raises
        ------
        TypeError
            If ``inertia_moment`` is not an instance of ``InertiaMoment``.

        See Also
        --------
        :py:class:`gearpy.units.units.InertiaMoment`
        """
        return super().inertia_moment

    @property
    def module(self) -> Optional[Length]:
        """Unit of the gear teeth size. It must be an instance of ``Length``. \n
        Once set at the spur gear instantiation, it cannot be changed afterward.

        Returns
        -------
        Length
            Unit of the gear teeth size.

        Raises
        ------
        TypeError
            If ``module`` is not an instance of ``Length``.

        See Also
        --------
        :py:class:`gearpy.units.units.Length`
        """
        return super().module

    @property
    def reference_diameter(self) -> Optional[Length]:
        """Reference diameter of the gear. It must be an instance of ``Length``. \n
        It is computed as the product of ``n_teeth`` times ``module`` at the spur gear instantiation and it cannot be
        changed afterward.

        Returns
        -------
        Length
            Reference diameter of the gear.

        See Also
        --------
        :py:class:`gearpy.units.units.Length`
        """
        return super().reference_diameter

    @property
    def face_width(self) -> Optional[Length]:
        """Face width of the gear. It must be an instance of ``Length``.

        Returns
        -------
        Length
            Face width of the gear.

        Raises
        ------
        TypeError
            If ``face_width`` is not an instance of ``Length``.

        See Also
        --------
        :py:class:`gearpy.units.units.Length`
        """
        return super().face_width

    @property
    def elastic_modulus(self) -> Optional[Stress]:
        """Elastic modulus of the material of the gear. It must be an instance of ``Stress``. It must be positive.

        Returns
        -------
        Stress
            Elastic modulus of the material of the gear.

        Raises
        ------
        TypeError
            If ``elastic_modulus`` is not an instance of ``Stress``.
        ValueError
            If ``elastic_modulus`` value is negative or null.

        See Also
        --------
        :py:class:`gearpy.units.units.Stress`
        """
        return super().elastic_modulus

    @property
    def lewis_factor(self) -> Optional[float]:
        """Factor used to compute stresses on the gear tooth. \n
        It is a tabular value that in general depends on the number of gear teeth and the pressure angle. In this
        case, the considered pressure angle is always 20 degrees, so the Lewis factor depends only on the number of gear
        teeth.

        Returns
        -------
        float
            Factor used to compute stresses on the gear tooth.
        """
        return self.__lewis_factor

    @property
    def driven_by(self) -> RotatingObject:
        """Rotating object that drives the gear, for example a motor, a flywheel or another gear. It must be a
        ``RotatingObject``.

        Returns
        -------
        RotatingObject
            Master rotating object that drives the gear.

        Raises
        ------
        TypeError
            If ``driven_by`` is not an instance of ``RotatingObject``.

        See Also
        --------
        :py:func:`gearpy.utils.relations.add_gear_mating`
        :py:func:`gearpy.utils.relations.add_fixed_joint`
        """
        return super().driven_by

    @driven_by.setter
    def driven_by(self, driven_by: RotatingObject):
        super(SpurGear, type(self)).driven_by.fset(self, driven_by)

    @property
    def drives(self) -> RotatingObject:
        """Rotating object driven by the gear, it can be a flywheel or another gear. It must be a ``RotatingObject``.

        Returns
        -------
        RotatingObject
            Rotating object driven by the gear.

        Raises
        ------
        TypeError
            If ``drives`` is not an instance of ``RotatingObject``.

        See Also
        --------
        :py:func:`gearpy.utils.relations.add_gear_mating`
        :py:func:`gearpy.utils.relations.add_fixed_joint`
        """
        return super().drives

    @drives.setter
    def drives(self, drives: RotatingObject):
        super(SpurGear, type(self)).drives.fset(self, drives)

    @property
    def master_gear_ratio(self) -> float:
        """Gear ratio of the mating between the gear and its driving gear. It must be a positive a float. \n
        If the gear is fixed to another driving ``RotatingObject``, then the ratio is ``1``, otherwise it is defined as
        the ratio between the gear number of teeth ``n_teeth`` and the same parameter of the master (driving) gear. \n
        To set this property use :py:func:`gearpy.utils.relations.add_gear_mating` or
        :py:func:`gearpy.utils.relations.add_fixed_joint`.

        Returns
        -------
        float
            Gear ratio of the mating between the gear and its driving gear.

        Raises
        ------
        TypeError
            If ``master_gear_ratio`` is not a float.
        ValueError
            If ``master_gear_ratio`` is negative or null.

        See Also
        --------
        :py:func:`gearpy.utils.relations.add_gear_mating`
        :py:func:`gearpy.utils.relations.add_fixed_joint`
        """
        return super().master_gear_ratio

    @master_gear_ratio.setter
    def master_gear_ratio(self, master_gear_ratio: float):
        super(SpurGear, type(self)).master_gear_ratio.fset(self, master_gear_ratio)

    @property
    def master_gear_efficiency(self) -> Union[float, int]:
        """Efficiency of the gear mating between the gear and its driving gear. It must be a float or an integer within
        ``0`` and ``1``. \n
        To set this property use :py:func:`gearpy.utils.relations.add_gear_mating` or
        :py:func:`gearpy.utils.relations.add_fixed_joint`.

        Returns
        -------
        float or int
            Efficiency of the gear mating between the gear and its driving gear.

        Raises
        ------
        TypeError
            If ``master_gear_efficiency`` is not a float or an integer.
        ValueError
            If ``master_gear_efficiency`` is not within ``0`` and ``1``.

        See Also
        --------
        :py:func:`gearpy.utils.relations.add_gear_mating`
        :py:func:`gearpy.utils.relations.add_fixed_joint`
        """
        return super().master_gear_efficiency

    @master_gear_efficiency.setter
    def master_gear_efficiency(self, master_gear_efficiency: Union[float, int]):
        super(SpurGear, type(self)).master_gear_efficiency.fset(self, master_gear_efficiency)

    @property
    def mating_role(self) -> Role:
        """Role of the gear in the gear mating. To set this parameter use ``add_gear_mating``. \n
        If the gear drives the mate one, then it is the "master" gear and its role is ``MatingMaster``, otherwise it is
        the "slave" one and its role is ``MatingSlave``.

        Returns
        -------
        Role
            The role of the gear in the gear mating.

        Raises
        ------
        ValueError
            If ``mating_role`` is not a subclass of ``Role``.
        """
        return super().mating_role

    @mating_role.setter
    def mating_role(self, mating_role: Role):
        super(SpurGear, type(self)).mating_role.fset(self, mating_role)

    @property
    def angular_position(self) -> AngularPosition:
        """Angular position of the gear. It must be an instance of ``AngularPosition``.

        Returns
        -------
        AngularPosition
            Angular position of the gear.

        Raises
        ------
        TypeError
            If ``angular_position`` is not an instance of ``AngularPosition``.

        See Also
        --------
        :py:class:`gearpy.units.units.AngularPosition`
        """
        return super().angular_position

    @angular_position.setter
    def angular_position(self, angular_position: AngularPosition):
        super(SpurGear, type(self)).angular_position.fset(self, angular_position)

    @property
    def angular_speed(self) -> AngularSpeed:
        """Angular speed of the gear. It must be an instance of ``AngularSpeed``.

        Returns
        -------
        AngularSpeed
            Angular speed of the gear.

        Raises
        ------
        TypeError
            If ``angular_speed`` is not an instance of ``AngularSpeed``.

        See Also
        --------
        :py:class:`gearpy.units.units.AngularSpeed`
        """
        return super().angular_speed

    @angular_speed.setter
    def angular_speed(self, angular_speed: AngularSpeed):
        super(SpurGear, type(self)).angular_speed.fset(self, angular_speed)

    @property
    def angular_acceleration(self) -> AngularAcceleration:
        """Angular acceleration of the gear. It must be an instance of ``AngularAcceleration``.

        Returns
        -------
        AngularAcceleration
            Angular acceleration of the gear.

        Raises
        ------
        TypeError
            If ``angular_acceleration`` is not an instance of ``AngularAcceleration``.

        See Also
        --------
        :py:class:`gearpy.units.units.AngularAcceleration`
        """
        return super().angular_acceleration

    @angular_acceleration.setter
    def angular_acceleration(self, angular_acceleration: AngularAcceleration):
        super(SpurGear, type(self)).angular_acceleration.fset(self, angular_acceleration)

    @property
    def torque(self) -> Torque:
        """Net torque applied on the gear. It must be an instance of ``Torque``. \n
        It is computed as the difference between ``driving_torque`` and ``load_torque``.

        Returns
        -------
        Torque
            Net torque applied on the gear.

        Raises
        ------
        TypeError
            If ``torque`` is not an instance of ``Torque``.

        See Also
        --------
        :py:class:`gearpy.units.units.Torque`
        """
        return super().torque

    @torque.setter
    def torque(self, torque: Torque):
        super(SpurGear, type(self)).torque.fset(self, torque)

    @property
    def driving_torque(self) -> Torque:
        """Driving torque applied on the gear by its driving gear. It must be an instance of ``Torque``.

        Returns
        -------
        Torque
            Driving torque applied on the gear by its driving gear.

        Raises
        ------
        TypeError
            If ``driving_torque`` is not an instance of ``Torque``.

        See Also
        --------
        :py:class:`gearpy.units.units.Torque`
        """
        return super().driving_torque

    @driving_torque.setter
    def driving_torque(self, driving_torque: Torque):
        super(SpurGear, type(self)).driving_torque.fset(self, driving_torque)

    @property
    def load_torque(self) -> Torque:
        """Load torque applied on the gear by its driven gear or an external load. It must be an instance of ``Torque``.

        Returns
        -------
        Torque
            Load torque applied on the gear by its driven gear or an external load.

        Raises
        ------
        TypeError
            If ``load_torque`` is not an instance of ``Torque``.

        See Also
        --------
        :py:class:`gearpy.units.units.Torque`
        """
        return super().load_torque

    @load_torque.setter
    def load_torque(self, load_torque: Torque):
        super(SpurGear, type(self)).load_torque.fset(self, load_torque)

    @property
    def tangential_force(self) -> Force:
        """Tangential force applied on the gear teeth by the mating gear. It must be an instance of ``Force``.

        Returns
        -------
        Force
            Tangential force applied on the gear teeth by the mating gear.

        Raises
        ------
        TypeError
            If ``tangential_force`` is not an instance of ``Force``.

        See Also
        --------
        :py:class:`gearpy.units.units.Force`
        :py:meth:`compute_tangential_force`
        """
        return super().tangential_force

    @tangential_force.setter
    def tangential_force(self, tangential_force: Force):
        super(SpurGear, type(self)).tangential_force.fset(self, tangential_force)

    def compute_tangential_force(self):
        """Computes the tangential force applied on the gear teeth by the mating gear. \n
        Considering a gear mating:

        - if the gear is the master one, then it takes into account the ``load_torque`` for the computation,
        - if the gear is the slave one, then it take into account the ``driving_torque`` for the computation.

        The tangential force is computed dividing the just described reference torque by the reference radius (half of
        the reference diameter).

        Raises
        ------
        ValueError
            If a gear mating between two gears has not been set.

        See Also
        --------
        :py:attr:`tangential_force`
        """
        if self.mating_role == MatingMaster:
            self.tangential_force = abs(self.load_torque)/(self.reference_diameter/2)
        elif self.mating_role == MatingSlave:
            self.tangential_force = abs(self.driving_torque)/(self.reference_diameter/2)
        else:
            raise ValueError("Gear mating not defined. "
                             "Use 'gearpy.utils.add_gear_mating' to set up a mating between two gears.")

    @property
    def tangential_force_is_computable(self) -> bool:
        """Whether is possible to compute the tangential force on the gear teeth. The tangential force computation
        depends on the value of ``module``, so if this optional parameter has been set at spur gear instantiation, then
        it is possible to compute the tangential force and this property is ``True``, otherwise is ``False``.

        Returns
        -------
        bool
            Whether is possible to compute the tangential force on the gear teeth.

        See Also
        --------
        :py:attr:`module`
        :py:attr:`tangential_force`
        :py:meth:`compute_tangential_force`
        """
        return super().tangential_force_is_computable

    @property
    def bending_stress(self) -> Stress:
        """Bending stress applied on the gear teeth by the mating gear. It must be an instance of ``Stress``.

        Returns
        -------
        Stress
            Bending stress applied on the gear teeth by the mating gear.

        Raises
        ------
        TypeError
            If ``bending_stress`` is not an instance of ``Stress``.

        See Also
        --------
        :py:class:`gearpy.units.units.Stress`
        :py:meth:`compute_bending_stress`
        """
        return super().bending_stress

    @bending_stress.setter
    def bending_stress(self, bending_stress: Stress):
        super(SpurGear, type(self)).bending_stress.fset(self, bending_stress)

    def compute_bending_stress(self):
        r"""Computes the bending stress applied on the gear teeth by the mating gear.

        Notes
        -----
        The bending stress computation is based on the following assumptions:

        - the tooth is stressed by the overall force acting on the tip of the tooth itself,
        - the most unfavorable situation is considered in the calculation, as if there is only one pair of teeth in
          contact within the contact segment,
        - the component of the overall force that determines the bending on the tooth is the only one considered and,
          for simplicity, is taken as having a value equal to the tangential force on the reference diameter,
        - the radial component of the overall force that causes a compressive stress on the tooth is neglected.

        The bending stress is computed with the following formula:

        .. math::
            \sigma_b = \frac{F_t}{m \, b \, Y_{LW}}

        where:

        - :math:`F_t` is the tangential force applied on the tooth,
        - :math:`m` is the gear module (``module``),
        - :math:`b` is the gear tooth face width (``face_width``),
        - :math:`Y_{LW}` is the gear Lewis factor (``lewis_factor``).

        See Also
        --------
        :py:attr:`bending_stress`
        :py:attr:`tangential_force`
        :py:attr:`module`
        :py:attr:`face_width`
        :py:attr:`lewis_factor`
        """
        self.bending_stress = self.tangential_force/(self.module*self.face_width)/self.lewis_factor

    @property
    def bending_stress_is_computable(self) -> bool:
        """Whether is possible to compute the bending stress on the gear teeth. The bending stress computation depends
        on the value of ``module`` and ``face_width``, so if these optional parameters have been set at spur gear
        instantiation, then it is possible to compute the bending stress and this property is ``True``, otherwise is
        ``False``.

        Returns
        -------
        bool
            Whether is possible to compute the bending stress on the gear teeth.

        See Also
        --------
        :py:attr:`module`
        :py:attr:`face_width`
        :py:attr:`bending_stress`
        :py:meth:`compute_bending_stress`
        """
        return super().bending_stress_is_computable

    @property
    def contact_stress(self) -> Stress:
        """Stress generated by the contact with mating gear teeth. It must be an instance of ``Stress``.

        Returns
        -------
        Stress
            The stress generated by the contact with mating gear teeth.

        Raises
        ------
        TypeError
            If ``contact_stress`` is not an instance of ``Stress``.

        See Also
        --------
        :py:class:`gearpy.units.units.Stress`
        :py:meth:`compute_contact_stress`
        """
        return super().contact_stress

    @contact_stress.setter
    def contact_stress(self, contact_stress: Stress):
        super(SpurGear, type(self)).contact_stress.fset(self, contact_stress)

    def compute_contact_stress(self):
        r"""Computes the contact stress generated by the contact with mating gear teeth.

        Raises
        ------
        ValueError
            - If a gear mating between two gears has not been set,
            - if gear role is ``MatingMaster`` and its slave gear misses ``module`` parameter,
            - if gear role is ``MatingMaster`` and its slave gear misses ``elastic_modulus`` parameter,
            - if gear role is ``MatingSlave`` and its master gear misses ``module`` parameter,
            - if gear role is ``MatingSlave`` and its master gear misses ``elastic_modulus`` parameter.

        Notes
        -----
        The contact stress computation is based on the following assumptions:

        - perfect elasticity of the materials the two mating gear,
        - absence of friction forces in the contact point,
        - small size of the contact surface compared to the size of the bodies between which contact occurs.

        The contact stress is computed with the following formula:

        .. math::
            \sigma_c = 0.262922 \, \sqrt{\frac{4 \, F_t}{b \, \cos \alpha \, \sin \alpha}
            \left( \frac{1}{D_1} + \frac{1}{D_2} \right) \, \frac{E_1 \, E_2}{E_1 + E_2}}

        where:

        - :math:`F_t` is the tangential force applied on the tooth,
        - :math:`b` is the gear face width (``face_width``),
        - :math:`\alpha` is the pressure angle of the gear, always equal to 20 degrees,
        - :math:`D_1` is the gear reference diameter (``reference_diameter``),
        - :math:`D_2` is the mating gear reference diameter,
        - :math:`E_1` is the gear elastic modulus (``elastic_modulus``),
        - :math:`E_2` is the mating gear elastic modulus.

        See Also
        --------
        :py:attr:`contact_stress`
        :py:attr:`tangential_force`
        :py:attr:`elastic_modulus`
        :py:attr:`face_width`
        :py:attr:`reference_diameter`
        """
        if self.mating_role == MatingMaster:
            if self.drives.module is not None:
                mate_reference_diameter = self.drives.reference_diameter
            else:
                raise ValueError(f"Impossible to compute contact stress for {self.__class__.__name__} {self.name!r} "
                                 f"because its mating {self.__class__.__name__} {self.drives.name!r} misses "
                                 f"'module' parameter.")
            if self.drives.elastic_modulus is not None:
                mate_elastic_modulus = self.drives.elastic_modulus
            else:
                raise ValueError(f"Impossible to compute contact stress for {self.__class__.__name__} {self.name!r} "
                                 f"because its mating {self.__class__.__name__} {self.drives.name!r} misses "
                                 f"'elastic_modulus' parameter.")
        elif self.mating_role == MatingSlave:
            if self.driven_by.module is not None:
                mate_reference_diameter = self.driven_by.reference_diameter
            else:
                raise ValueError(f"Impossible to compute contact stress for {self.__class__.__name__} {self.name!r} "
                                 f"because its mating {self.__class__.__name__} {self.driven_by.name!r} misses "
                                 f"'module' parameter.")
            if self.driven_by.elastic_modulus is not None:
                mate_elastic_modulus = self.driven_by.elastic_modulus
            else:
                raise ValueError(f"Impossible to compute contact stress for {self.__class__.__name__} {self.name!r} "
                                 f"because its mating {self.__class__.__name__} {self.driven_by.name!r} misses "
                                 f"'elastic_modulus' parameter.")
        else:
            raise ValueError("Gear mating not defined. "
                             "Use 'gearpy.utils.add_gear_mating' to set up a mating between two gears.")

        equivalent_elastic_modulus = 2*self.elastic_modulus* \
                                     (mate_elastic_modulus/(self.elastic_modulus + mate_elastic_modulus))
        inverse_curvature_sum = self.__PRESSURE_ANGLE.sin()/2*self.reference_diameter* \
                                (mate_reference_diameter/(self.reference_diameter + mate_reference_diameter))
        contact_pressure = self.tangential_force/self.__PRESSURE_ANGLE.cos()/(self.face_width*inverse_curvature_sum)

        self.contact_stress = Stress(value = 0.262922*sqrt(equivalent_elastic_modulus.to('Pa').value*
                                                           contact_pressure.to('Pa').value),
                                     unit = 'Pa')

    @property
    def contact_stress_is_computable(self) -> bool:
        """Whether is possible to compute the contact stress force on the gear teeth. The contact stress computation
        depends on the value of ``module``, ``face_width`` and ``elastic_modulus``, so if these optional parameters have
        been set at spur gear instantiation, then it is possible to compute the contact stress and this property is
        ``True``, otherwise is ``False``.

        Returns
        -------
        bool
            Whether is possible to compute the contact stress on the gear teeth.

        See Also
        --------
        :py:attr:`module`
        :py:attr:`face_width`
        :py:attr:`elastic_modulus`
        :py:attr:`contact_stress`
        :py:meth:`compute_contact_stress`
        """
        return super().contact_stress_is_computable

    @property
    def external_torque(self) -> Callable[[AngularPosition, AngularSpeed, Time], Torque]:
        """Custom function to compute the external torque applied on the gear. It must be a function with parameters
        ``angular_position``, ``angular_speed`` and ``time``. The function must return an instance of ``Torque``.

        Returns
        -------
        Callable
            The function to compute the external torque applied on the gear.

        Raises
        ------
        TypeError
            If ``external_torque`` is not callable.
        KeyError
            If ``external_torque`` misses parameters ``angular_position``, ``angular_speed`` or ``time``.

        Examples
        --------
        Constant torque, not dependent on ``angular_position``, ``angular_speed`` or ``time``.

        >>> from gearpy.mechanical_objects import SpurGear
        >>> from gearpy.units import InertiaMoment, Torque
        >>> gear = SpurGear(name = 'gear', n_teeth = 10, inertia_moment = InertiaMoment(1, 'kgm^2'))
        >>> gear.external_torque = lambda angular_position, angular_speed, time: Torque(5, 'Nm')

        Torque dependent on ``angular_position`` and ``time``. \n
        In this case the gear gets a periodic load, dependent on time, and an extra load dependent on its angular
        position. The dependence by angular position may be used to model cases where cams are involved.

        >>> import numpy as np
        >>> from gearpy.units import AngularPosition, AngularSpeed, Time
        >>> def custom_external_torque(angular_position: AngularPosition,
        ...                            angular_speed: AngularSpeed,
        ...                            time: Time):
        ...     return Torque(value = angular_position.sin() +
        ...                           np.cos(time.to('sec').value),
        ...                   unit = 'Nm')
        >>> gear.external_torque = custom_external_torque

        Torque dependent on ``angular_position``, ``angular_speed`` and ``time``. \n
        With respect ot the previous case, the gear gets an extra load dependent on its angular speed. The dependence by
        angular speed may be used to model cases where air friction is not negligible.

        >>> def complex_external_torque(angular_position: AngularPosition,
        ...                             angular_speed: AngularSpeed,
        ...                             time: Time):
        ...     return Torque(value = angular_position.sin() +
        ...                           0.001*(angular_speed.to('rad/s').value)**2 +
        ...                           np.cos(time.to('sec').value),
        ...                   unit = 'Nm')
        >>> gear.external_torque = complex_external_torque
        """
        return super().external_torque

    @external_torque.setter
    def external_torque(self, external_torque: Callable[[AngularPosition, AngularSpeed, Time], Torque]):
        super(SpurGear, type(self)).external_torque.fset(self, external_torque)

    @property
    def time_variables(self) -> Dict[str, List[UnitBase]]:
        """Time variables of the gear. Each time variable is stored as a dictionary key-value pair. The available time
        variables are:

        - ``angular position``,
        - ``angular speed``,
        - ``angular acceleration``,
        - ``torque``,
        - ``driving torque``,
        - ``load torque``,
        - ``tangential force``,
        - ``bending stress``,
        - ``contact stress``.

        ``tangential force``, ``bending stress`` and ``contact stress`` are listed among time variables only if they are
        computable indeed, depending on which gear parameters are set at gear instantiation. \n
        Corresponding values of the dictionary are lists of the respective time variable values. \n
        At each time iteration, the ``Solver`` appends every time variables' values to the relative list in the
        dictionary.

        Returns
        -------
        dict
            Time variables of the gear.

        See Also
        --------
        :py:meth:`update_time_variables`
        """
        return super().time_variables

    def update_time_variables(self) -> None:
        """Updates ``time_variables`` dictionary by appending the last value of each time variable (key of the
        dictionary) to corresponding list (value of the dictionary).

        See Also
        --------
        :py:attr:`time_variables`
        """
        super().update_time_variables()
        if self.tangential_force_is_computable:
            self.time_variables['tangential force'].append(self.tangential_force)

            if self.bending_stress_is_computable:
                self.time_variables['bending stress'].append(self.bending_stress)

                if self.contact_stress_is_computable:
                    self.time_variables['contact stress'].append(self.contact_stress)
